@@ -1,13 +1,7 @@
-import itertools
-import logging
-import os
-import traceback
-from urllib import urlencode
-from urlparse import urlparse
-
+# -*- coding: utf-8 -*-
 from AccessControl import getSecurityManager
 from Acquisition import aq_parent
-from Products.CMFCore.utils import getToolByName
+from BTrees.OOBTree import OOBTree
 from lxml import etree
 from lxml import html
 from plone.keyring.interfaces import IKeyManager
@@ -17,33 +11,63 @@ from plone.protect.authenticator import createToken
 from plone.protect.authenticator import isAnonymousUser
 from plone.protect.interfaces import IConfirmView
 from plone.protect.interfaces import IDisableCSRFProtection
-from plone.protect.utils import SAFE_WRITE_KEY
 from plone.protect.utils import getRoot
 from plone.protect.utils import getRootKeyManager
+from plone.protect.utils import SAFE_WRITE_KEY
 from plone.protect.utils import safeWrite  # noqa b/w compat import
 from plone.transformchain.interfaces import ITransform
+from Products.CMFCore.utils import getToolByName
 from repoze.xmliter.serializer import XMLSerializer
 from repoze.xmliter.utils import getHTMLSerializer
+from six.moves.urllib.parse import urlencode
+from six.moves.urllib.parse import urlparse
+from zExceptions import Forbidden
+from zope.component import adapts
+from zope.component import ComponentLookupError
+from zope.component import getUtility
+from zope.interface import implementer
+from zope.interface import Interface
+
+import itertools
+import logging
+import os
+import pkg_resources
+import six
+import traceback
 import transaction
 import types
-from zExceptions import Forbidden
-from zope.component import ComponentLookupError
-from zope.component import adapts
-from zope.component import getUtility
-from zope.interface import implementer, Interface
-from zope.globalrequest import getRequest
+
 
 try:
     from zope.component.hooks import getSite
 except:
     from zope.app.component.hooks import getSite
 
+try:
+    pkg_resources.get_distribution('plone.app.blob')
+except pkg_resources.DistributionNotFound:
+    ATBlob = None
+else:
+    from plone.app.blob.content import ATBlob
 
-LOGGER = logging.getLogger('plone.protect')
+try:
+    from plone.scale.storage import ScalesDict
+except ImportError:
+    ScalesDict = None
+
+
+logger = logging.getLogger('plone.protect')
 
 X_FRAME_OPTIONS = os.environ.get('PLONE_X_FRAME_OPTIONS', 'SAMEORIGIN')
 CSRF_DISABLED = os.environ.get('PLONE_CSRF_DISABLED', 'false').lower() in \
     ('true', 't', 'yes', 'y', '1')
+ANNOTATION_KEYS = (
+    'plone.contentrules.localassignments',
+    'syndication_settings',
+    'plone.portlets.contextassignments',
+    'plone.scale',
+)
+SAFE_TYPES = tuple(t for t in [ATBlob, ScalesDict] if t is not None)
 
 
 @implementer(ITransform)
@@ -60,6 +84,9 @@ class ProtectTransform(object):
 
     key_manager = None
     site = None
+    safe_views = (
+        'plone_lock_operations',
+    )
 
     def __init__(self, published, request):
         self.published = published
@@ -67,7 +94,8 @@ class ProtectTransform(object):
 
     def parseTree(self, result, encoding):
         # if it's a redirect, the result shall not be transformed
-        request = getRequest()
+        request = self.request
+
         if request.response.status in (301, 302):
             return None
 
@@ -85,6 +113,10 @@ class ProtectTransform(object):
                                                    'compress',):
             return None
 
+        if isinstance(result, list) and len(result) == 1:
+            # do not parse empty strings to omit warning log message
+            if not result[0].strip():
+                return None
         try:
             result = getHTMLSerializer(
                 result, pretty_print=False, encoding=encoding)
@@ -92,14 +124,14 @@ class ProtectTransform(object):
             # output does odd character encodings
             result.serializer = html.tostring
             return result
-        except (TypeError, etree.ParseError):
+        except (AttributeError, TypeError, etree.ParseError):
             # XXX handle something special?
-            LOGGER.warn('error parsing dom, failure to add csrf '
+            logger.warn('error parsing dom, failure to add csrf '
                         'token to response for url %s' % self.request.URL)
             return None
 
     def transformBytes(self, result, encoding):
-        result = unicode(result, encoding, 'ignore')
+        result = six.text_type(result, encoding, 'ignore')
         return self.transformIterable([result], encoding)
 
     def transformString(self, result, encoding):
@@ -111,9 +143,11 @@ class ProtectTransform(object):
     def transformIterable(self, result, encoding):
         """Apply the transform if required
         """
-
         # before anything, do the clickjacking protection
-        if X_FRAME_OPTIONS and not self.request.response.getHeader('X-Frame-Options'):
+        if (
+            X_FRAME_OPTIONS and
+            not self.request.response.getHeader('X-Frame-Options')
+        ):
             self.request.response.setHeader('X-Frame-Options', X_FRAME_OPTIONS)
 
         if CSRF_DISABLED:
@@ -167,6 +201,12 @@ class ProtectTransform(object):
             return published.im_self
         return aq_parent(published)
 
+    def getViewName(self):
+        try:
+            return self.getContext().__name__
+        except AttributeError:
+            return None
+
     def check(self):
         """
         just being very careful here about our check so we don't
@@ -176,7 +216,7 @@ class ProtectTransform(object):
             return self._check()
         except:
             transaction.abort()
-            LOGGER.error("Error checking for CSRF. "
+            logger.error("Error checking for CSRF. "
                          "Transaction will be aborted since the request "
                          "is now unsafe:\n%s" % (
                              traceback.format_exc()))
@@ -188,19 +228,22 @@ class ProtectTransform(object):
             conn._registered_objects
             # skip the 'temporary' connection since it stores session objects
             # which get written all the time
-            for name, conn in app._p_jar.connections.items() if name != 'temporary'
+            for name, conn in app._p_jar.connections.items()
+            if name != 'temporary'
         ]))
 
     def _check(self):
         registered = self._registered_objects()
         if len(registered) > 0 and \
                 not IDisableCSRFProtection.providedBy(self.request):
+            if self.getViewName() in self.safe_views:
+                return True
             # Okay, we're writing here, we need to protect!
             try:
                 check(self.request, manager=self.key_manager)
                 return True
             except ComponentLookupError:
-                LOGGER.info('Can not find key manager for CSRF protection. '
+                logger.info('Can not find key manager for CSRF protection. '
                             'This should not happen.')
                 raise
             except Forbidden:
@@ -216,16 +259,36 @@ class ProtectTransform(object):
                     safe_oids = self.request.environ[SAFE_WRITE_KEY]
                 safe = True
                 for obj in registered:
-                    if (not IPortletAssignment.providedBy(obj) and
-                            getattr(obj, '_p_oid', False) not in safe_oids):
+                    if IPortletAssignment.providedBy(obj):
+                        continue
+                    if getattr(obj, '_p_oid', False) in safe_oids:
+                        continue
+                    if SAFE_TYPES and isinstance(obj, SAFE_TYPES):
+                        continue
+                    if isinstance(obj, OOBTree):
                         safe = False
-                        break
+                        for key in ANNOTATION_KEYS:
+                            try:
+                                if key in obj:
+                                    safe = True
+                                    break
+                            except TypeError:
+                                pass
+                        if safe:
+                            continue
+                    safe = False
+                    break
                 if not safe:
                     if self.request.REQUEST_METHOD != 'GET':
                         # only try to be "smart" with GET requests
                         raise
-                    LOGGER.info('%s\naborting transaction due to no CSRF '
-                                'protection on url %s'%(traceback.print_stack(), self.request.URL))
+                    logger.info(
+                        '{0:s}\naborting transaction due to no CSRF '
+                        'protection on url {1:s}'.format(
+                            '\n'.join(traceback.format_stack()),
+                            self.request.URL
+                        )
+                    )
                     transaction.abort()
 
                     # conditions for doing the confirm form are:
@@ -260,6 +323,7 @@ class ProtectTransform(object):
         return True
 
     def transform(self, result, encoding):
+
         result = self.parseTree(result, encoding)
         if result is None:
             return None
@@ -269,7 +333,10 @@ class ProtectTransform(object):
             token = createToken(manager=self.key_manager)
         except ComponentLookupError:
             if self.site is not None:
-                LOGGER.warn('Keyring not found on site. This should not happen', exc_info=True)
+                logger.warn(
+                    'Keyring not found on site. This should not happen',
+                    exc_info=True
+                )
             return result
 
         for form in root.cssselect('form'):
@@ -300,15 +367,17 @@ class ProtectTransform(object):
         if self.site is not None and not root.cssselect('#protect-script'):
             # Alternative: add this in the resource registry.
             site_url = self.site.absolute_url()
-            body = root.cssselect('body')[0]
-            protect_script = etree.Element("script")
-            protect_script.attrib.update({
-                'type': "application/javascript",
-                'src': "%s/++resource++protect.js" % site_url,
-                'data-site-url': site_url,
-                'data-token': token,
-                'id': 'protect-script'
-            })
-            body.append(protect_script)
+            elements = root.cssselect('body')
+            if len(elements):
+                body = elements[0]
+                protect_script = etree.Element("script")
+                protect_script.attrib.update({
+                    'type': "application/javascript",
+                    'src': "%s/++resource++protect.js" % site_url,
+                    'data-site-url': site_url,
+                    'data-token': token,
+                    'id': 'protect-script'
+                })
+                body.append(protect_script)
 
         return result
